\     Turtle-Logo in Forth, Copyright (c) 2016-2018, Assad Ebrahim, Mathematical Science & Technologies
\          FREE download from: http://www.mathscitech.org/articles/turtle-logo-forth
\          Email assad.ebrahim@alum.swarthmore.edu for comments, bug-reports, feature-requests

\     GNU License - Users are permitted free re-use provided:                          
\           1/ Source code for all modified versions are made freely available under the GNU License            
\           2/ Attribution and this header is preserved (lines 1-7)

\     Description: This is a Turtle logo program for Windows operated entirely with a few keys.  
\                  It has several advanced features, including macro recording and replay.
\                  It is written in ~800 lines of Forth code for the F-PC Forth for DOS.
\                  To run it requires a DOS emulator, either vDOSplus (recommended) or DOSbox (both freely available)
\                  A complete run-time package with Windows installer is available at the website listed above.  

\     Turnkey Startup: Run vDOSplus.exe which in turn calls "turtle_logo.bat".
\                      Further instructions are provided in the _readme.txt

\     Developers Startup: To modify the program, edit turtle.fs and run it in F-PC (Forth FPC) using command:
\                                   fload turtle.fs, then: turtle-go
\                        To package a modified release, recompile F-PC with FMETA to create F-PC with turtle.fs embedded in it (turnkey application)
\                        F-PC can be downloaded from: ftp://ftp.taygeta.com/pub/Forth/Compilers/native/dos/FPC/fpc36.zip
\                        Instructions included in docs: Fpc\doc\ting_1989_FPC_UsersManual\Chapter9_Rebuilding_FPC.txt


\ Revision History
: rev# ." 1.5.4023" ;
: welcome ( -- )
       ." Welcome to Turtle Logo! (" rev# ." , (c) 2016-2018, Assad Ebrahim)"
    cr ." QUICK START:"
    cr ."   (1) To MOVE Turtle, use the arrow keys and backspace..."
    cr ."     FWD moves one step along current heading, and paints."
    cr ."     TURN LEFT/RIGHT changes Turtle's heading 45 deg counter-/clock- wise"
    cr ."     BACKSPACE backs up and erases."                   \ Note: movement wraps around the edges of the screen to the other side (Exercise)
    cr ."   (2) To CHANGE COLOR... "
    cr ."     C / X cycles pen color (fwd/back) through 16 choices."
    cr ."     ' toggles pen up (to move without painting) or down (to draw again)"
    cr ."     H sends Turtle to home position at bottom left of screen."
    cr ."     ; toggles Turtle's visibility (hide or show)."                            \ instruction not on main screen
    cr ."   (3) To TEACH Turtle some tricks, you can have him learn and repeat some steps..."  \ RECORD/REPLAY macros..."
    cr ."     T to begin/end TEACHING Turtle a particular SKILL. (Turtle can learn 9 skills.)"       \ skill=diary.  can record nested macros, but not recursive macros
    cr ."     N cycles through skills 1-9.  Pressing 1-9 gets Turtle to replay that skill."     \ Note, it will be EXACTLY what you told him. \ Later add LOOP
    cr ."   (4) Other actions..."
    cr ."     S saves the screen; U undoes all actions to the last save point
    cr ."     ESC clears screen -- once you confirm."
    cr ."     ? shows this HELP screen again (your drawing is safe!)"
    cr ."     Q quits the program (after confirming)"
    cr ." Happy Turtle-ing!  
    cr
    cr ." For ideas & challenge problems, see http://www.mathscitech.org/articles/turtle-logo-forth"
    cr ;


\ === Interactivity / Scaffolding
$1B          constant    esc-key                 \ 27d
: pause-key ( -- ) ." PRESS ANY KEY TO CONTINUE... " key drop ;
: pause-quit ( -- ) key 'q' over = swap 'Q' = or if abort then ;
: pq ( -- ) pause-quit ;   \ abbreviation
: prompt-clear ( -- ) ." Clear screen? (Press ESC to confirm, or any key to continue drawing.)" ; 
: clear? ( -- t/f )   \ y,Y,ESC are confirmation
    prompt-clear 'y' 'Y' esc-key 
    key >r r@ = swap r@ = or swap r> = or               \ check key against queue of acceptable values
    if true else false then ;
: prompt-quit ( -- ) ." ...sure? (y or q to really quit, else any key to return.)"  ;
: quit? ( -- t/f )   \ q,Q,y,Y are confirmation
    prompt-quit 'y' 'Y' 'q' 'Q'
    key >r r@ = swap r@ = or swap r@ = or swap r> = or  \ check key against queue of acceptable values
    if true else false then ;
: arrow ( -- )  '>' '-' emit emit ;
: .key ( asc -- ) dup . arrow space emit ;

\ === Numeric conversion
: a2d ( asc -- digit )  '0' - ;   \ ascii to digit
: d2a ( digit -- asc )  '0' + ;   \ digit to ascii

\ === Memory
: bytes     ( n -- n )  ;                                 \ 1x, for documentation
: cells     ( num -- num*wsz )  2 *  ;                    \ 16-bit word length on 8086 chip
: c@+       ( adr -- adr+ val )   dup 1 + swap c@ ;       \ fetch char and advance
: @+        ( adr -- adr+ val )   dup 1 cells + swap @ ;  \ fetch word and advance
: c!+       ( val adr -- adr+ )   dup 1 + -rot c! ;       \ store char and advance
: !+        ( val adr -- adr+ )   dup 1 cells + -rot ! ;  \ store word and advance

\ === Strings
$22   constant    dquote    \ " double quote
: s" ( | <wordstream"> -- c-addr nchar  ) dquote word c@+ pad swap 2dup >r >r cmove r> r> ;  \ take string from wordstream input
: cmove+ ( c-addr ptr n -- ptr' )  >r tuck r@ cmove r> + ;  \ copy byte sequence/string (c-addr, n bytes) to ptr, returning new position
: cmove-fwBL+ ( adr1 adr2 n fwlen -- adr' )                 \ copy adr1 to adr2 (n bytes) padding rest with BL (fixed width field)
    >r >r tuck r@ cmove r>   \ copy str to array
    2dup + ( array n array' ) 
    r@ rot - BL fill                          \ pad rest of fixed width field with blanks
    r> + ;                                    \ advance array    
: attach-string-fw ( d-adr fwlen str n -- adr' fwlen ) >r rot r> 3 pick cmove-fwBL+ swap ;   \ attach a fixed width string to an array
: get-str ( index array strlen -- ptr len ) rot over * rot + swap ;  \ Returns string ptr/len at index when working with an array of fixed length strings

\ === Color encoding (Text mode): colorcode hi nybble holds background color, low nybble holds foreground color
: shl4      ( val -- val )        $10 * ;                 \ raise nybble
: shr4      ( val -- val )        $10 / ;                 \ lower nybble
: fgbg2col ( fgcolor bgcolor -- colorcode )  shl4 + ;
: col2fgbg ( colorcode -- fgcolor bgcolor )  dup $0F and  ( fgcolor )  swap shr4  ( bgcolor ) ;

\ =========  Colors
16               constant   num-colors
num-colors       constant   penup-col-index
num-colors 1 +   constant   num-color-str    \ 1 more for penup string
 0      constant  black
 1      constant  navy
 2      constant  green
 3      constant  teal  
 4      constant  red   
 5      constant  purple 
 6      constant  brown 
 7      constant  silver  
 8      constant  slate 
 9      constant  blue  
 10     constant  lime  
 11     constant  aqua  
 12     constant  orange
 13     constant  pink  
 14     constant  yellow
 15     constant  white 
 silver constant  grey     \ alternate spellings
 silver constant  gray
 slate  constant  darkgrey
 slate  constant  darkgray

 \ ===  Color Name Strings
6        constant    strlen-col
: attach-colors ( c-addr L n | <wordstream> -- )   0 do BL word c@+ attach-string-fw loop 2drop ; \ attaches words from input stream as fixed width words
: create-color-names                         \ data structure (object) holding n colornames as fixed-width strings (L=strlen-col)
create  ( L n | <name> <wordstream> --  )    \ colornames pulled from wordstream
    here -rot                                \ store ptr to object
    2dup * 2 + bytes allot ( addr L n )      \ allocate storage (strings+metadata)
    rot >r 2dup swap r> c!+ c!+              \ metadata L n stored at start of object
    -rot ( c-addr L n )                      
    attach-colors                            \ strings copied from wordstream
does> ( index -- c-addr nchar )              \ get fixed width string at index i
    c@+ -rot 1 + -rot ( c-adr L index )      \ extract fw-size and get past metadata
    over * rot + swap ;                      \ return specified string
\ Color comments: navy=dark blue, teal=aqua, red=maroon, silver=grey, slate=darkgrey, blue=lt blue, lime=lt green, aqua=lt aqua/cyan, pink=magenta/fuschia
strlen-col num-color-str CREATE-COLOR-NAMES color-names Black Navy Green Teal Red Purple Brown Silver Slate Blue Lime Aqua Orange Pink Yellow White PEN-UP

create screen-color 1 bytes allot                   \ VARIABLE - holds canvas color codes 0..15
: set-screen-color ( color -- ) screen-color c! ;
: get-screen-color ( -- color ) screen-color c@ ;

create cursor-color 1 bytes allot                   \ VARIABLE - holds cursor color (bg color)
: set-cursor-color ( bgcolor -- ) cursor-color c! ;
: get-cursor-color ( -- bgcolor ) cursor-color c@ ;

\ === Pen/Cursor color mapping (foreground/background).
\ The design chooses an optimal cursor color (bgcolor) for each pen-color (fgcolor) to maximize their contrast ratio (see Table in documentation)
\ Rationale: We want color of pen (fgcolor) to contrast sufficiently with background to discern.  
\ Method: We use theory of color contrasts to choose the optimal pairing (equivalent to designer-provided) using color contrast checker tool (contrast ratio)
\ References:   http://www.worqx.com/color/color_contrast.htm
\               http://webaim.org/resources/contrastchecker/
\               http://juicystudio.com/services/luminositycontrastratio.php#specify
\ (ALTERNATIVE: could have it be paint color, then would need fg char to be high contrast or flashing)

\ fgbg lookup table is indexed by fgcolor and holds bgcolor (cursor) to give the optimal combination 
: create-fgbg-lookup     \ for optimized color palette
create     \ store bg color into table indexed by fg  (first column is shown for readability only; order must be this way for color to work as index)
    \ fg    bg
    black    silver  c, drop
    navy     aqua    c, drop
    green    lime    c, drop   \ not great, but ok
    teal     aqua    c, drop
    red      yellow  c, drop
    purple   white   c, drop
    brown    silver  c, drop
    silver   black   c, drop
    slate    silver  c, drop
    blue     aqua    c, drop
    lime     slate   c, drop
    aqua     navy    c, drop
    orange   yellow  c, drop
    pink     black   c, drop
    yellow   brown   c, drop
    white    black   c, drop 
\ Note: penup is handled with specific penup color instructions
does> ( fg -- bg ) + c@ set-cursor-color ;   \ get-optim-big-color
CREATE-FGBG-LOOKUP  set-optim-bg-color

create pen-color 1 bytes allot  \ VARIABLE -- pencolor is foreground color, 0..15
: set-pen-color ( c -- )  dup   pen-color c! set-optim-bg-color ;      \ automatically change background so turtle can be seen better
: set-penup-color ( -- )  white pen-color c! lime set-cursor-color ;   \ gives penup a distinct appearance
: get-pen-color ( -- c )  pen-color c@ ;

create prev-pen-color 1 bytes allot  \ VARIABLE
: save-prev-pen-color    ( -- )   get-pen-color prev-pen-color c! ;
: restore-prev-pen-color ( -- c ) prev-pen-color c@ set-pen-color ;  \ ??? TESTER REVIEW search previous comment

create border-color 1 bytes allot 
: set-border-color ( color -- ) border-color c! ;
: get-border-color ( -- color ) border-color c@ ;

create penup-state  1 bytes allot  \ VARIABLE -- BOOLEAN true if pen is up else false
: set-penup-state   ( t/f -- ) penup-state c! ; 
: get-penup-state   ( -- ) penup-state c@ ;

create rec-state 1 bytes allot \ States describing status of Recording (on or off or full etc.)
: set-rec-state ( r -- ) rec-state c! ;
: get-rec-state ( -- r ) rec-state c@ ;
0 constant rec-state-ready
1 constant rec-state-recording
2 constant rec-state-full
3 constant rec-state-reminder  \ still recording, i.e. a flash
4 constant rec-state-paused

\ ??? CODE REVIEW --- this should be re-written using Create..does>
\ Record strings
5       constant   nrec-msg                                \ number of different recording messages
12      constant   strlen-rec                              \ msg length
create  rec-msg-names    nrec-msg  strlen-rec  *  bytes  allot   \ VARIABLE recording string
: clear-rec-msg-names ( -- ) rec-msg-names nrec-msg strlen-rec * BL fill ;
clear-rec-msg-names
: show-rec-msg-names ( -- ) rec-msg-names nrec-msg strlen-rec * type cr pq ;
: attach-rec-msgs ( -- )  
    rec-msg-names strlen-rec
    " TeachSkill 1"     attach-string-fw   \ #1 is default diary/skill, update-diary-string keeps these current and updates menu dynamically
    " LEARNING...1"     attach-string-fw
    "  Full!"           attach-string-fw
    " LEARNING...1"     attach-string-fw 
    " PAUSED...1"     attach-string-fw 
    2drop ;
: get-rec-str ( -- ptr n ) get-rec-state rec-msg-names strlen-rec get-str ; 

11 constant diary-num-pos   \ diary num position in string
: update-diary-string ( digit -- ) 1 + d2a rec-msg-names diary-num-pos + ( val ptr ) 2dup c!  \ first string
        strlen-rec + 2dup c!             \ second string
        strlen-rec + strlen-rec + c! ;   \ fourth string


\ === Heading through 8 Compass Directions
\ Heading is an integer 0 - 7 which indexes into a state array and gets the heading character
\ The maths of rotation is mod, so this is arithmetic mod-8
\ Turtle cycles through the directions as:
\   cw:  N > NE > E > SE > S > SW > W > NW
\   ccw: opposite

create heading 1 bytes allot  \ 0 to 7    \  VARIABLE
: set-heading ( h# -- ) heading c! ;
: init-heading ( -- ) 0 set-heading ;  \ default heading is N=0
: get-heading ( -- h# ) heading c@ ;

\ graphical symbols to show what direction turtle is heading
: create-heading-symbol
create ( <name> -- ) '^' c, '/' c, '>' c, '\' c, 'v' c, '/' c, '<' c, '\' c,   \ headings in cyclic order (N,NE,E,...)
does> ( h# -- asc )  + c@ ;
CREATE-HEADING-SYMBOL  get-heading-symbol

\ direction names, letters backwards because numbers are stored in lower-endian form, i.e. low byte first
: create-heading-name
create ( <name> -- )
        'N' c,  BL c,
        'N' c, 'E' c, 
        'E' c,  BL c, 
        'S' c, 'E' c,
        'S' c,  BL c,
        'S' c, 'W' c,
        'W' c,  BL c,
        'N' c, 'W' c,
does> ( h# -- ptr ) swap cells + ;
CREATE-HEADING-NAME  get-heading-string


\ =========== Graphics.
\ VGA in DOS is a quick way to produce graphics because in DOS it is easy to access screen pixels (peek/poke) directly via VGA card memory.  
\ This is no \ longer the case in modern windows which requires going through a graphics library such as SDL or OpenGL.
\ This program is written F-PC (16-bit Forth) because F-PC makes it easy to write 8086 assembly code, making VGA on DOS straightforward. 
\ Modern Windows computer?  No problem, use a DOS emulator (vDosPlus recommended or DOSBox).

\ Standard text mode graphics works with 80 columns and 25 rows (2000 cells, indices x:0..79, y:0..24 ).  
\ Implementation of each row is with 160 data columns, so 4000 cells.  
\ Drawing works correctly only on even columns so in practice the 80 column indices are  multiples of 2, i.e. 0,2,4,...,158.
\ Pixel locations are specified in linear order from 0 to 3998d ($F9E in hex) with the final 3999d cell not visible (even pixel only issue).

\ Two representations are possible: modified Cartesian x,y as across and *down* vs. Matrix: i,j for row and column (down/across).
\ THIS PROGRAM uses Cartesian (x,y / across/down).

\ We have THREE coordinate systems: 
\     USER COORDINATES - what the user can access on the screen, i.e. excluding menus, borders, etc., with 0,0 as the origin point for the user's cursor.
\           i.e. within a window, (i.e. a subset of the 80x25 full-screen)
\     SCREEN COORDINATES - where this all happens on the screen, with (0,0) at top left corner of the screen
\     PHYSICAL COORDINATES - the underlying implementation / linear data array against which the drawing pixels must be referenced.
\   xy2lin converts from USER coordinates from physical position in linear array.  lin2xy goes from physical to user.
\   xy2lin-screen converts from SCREEN coordinates to physical.  No need for back-conversion.

\ Example - FPC uses rows 0-1 (1st line is Forth bar, and 2nd line is erased after command scrolls screen up), so to see a pixel within F-PC start at linear position 320 (start of row 3)
\ FPC uses columns 69-79 of the first 12 rows for dictionary, with the full 80 cols starting from row 13.

\   References:
\   Video programming:  http://ece.wpi.edu/~wrm/Courses/EE3803/Labs/roehrl.html
\   Video mode:         http://www.wagemakers.be/english/doc/vga

\ PC Video Mode (BIOS settings)
$1 constant text40x25-8color
$3 constant text80x25-8color            \ text mode (8 color)
$8 constant text80x25-16color           \ default F-PC mode, colsxrows, each character is 8x8 pixels (640 pixels x 200)
$4 constant graphCGA-320x200-8color     \ standard resolution
$10 constant graphEGA-640x350-16color   \ high resolution
$12 constant graphVGA-640x480-16color   \ default VGA graphics mode

code set-vid-mode ( mode -- )   \ calls PC BIOS instruction to set desired video mode
    pop ax          \ AL=video mode 
    mov ah, # 0     \ AH=set video mode (BIOS instruction)
    int $10		    \ DOS interrupt: call BIOS instruction
next end-code

: set-video-mode ( -- ) text80x25-16color set-vid-mode ;
\ : set-video-mode ( -- ) graphVGA-640x480-16color set-vid-mode ;

\ === Physical (linear array) coordinates for text mode: 80x25 = 2000 visible cells.  160x25 = 4000 implementation cells.
  80  constant screen-cols
 160  constant physical-cols  \ note: odd slots are invalid display locations, so counter runs 0,2,4,..,178   (see graphics discussion above)
  25  constant screen-rows
   0  constant physical-zero
3998  constant physical-end    \ $F9E hex
\ From PHYSICAL to SCREEN coordinates
 1  constant top-menu-height
 1  constant bottom-menu-height
 1  constant buffer-thickness  \ frame all round

\ === USER area in SCREEN coordinates (0,0 is top left pixel of the screen)
 0 top-menu-height + buffer-thickness +                  constant screen-top
 screen-rows bottom-menu-height -  buffer-thickness - 1 -   constant screen-bot
 buffer-thickness                                        constant screen-left
 screen-cols buffer-thickness - 1 -                         constant screen-right
screen-bot screen-top - 1 +                              constant screen-height
screen-right screen-left - 1 +                           constant screen-width
 3 1 - physical-cols *                                       constant row#3             \ USER COORDINATE STARTS HERE 3rd line ($140=320d) is one accessible in  user mode (1st is menu, 2nd is border)

\ === Linear offset in physical coordinates to 0,0 in user coordinates (top left pixel that the user can access, i.e. excluding menus and buffers)
 create lin-offset 1 cells allot  \ VARIABLE
: set-lin-offset ( -- ) screen-top physical-cols * screen-left 2 * + lin-offset ! ;
: get-lin-offset ( -- offset ) lin-offset @ ;

: set-interactive-graphics-mode ( -- )  row#3 lin-offset ! ;
set-interactive-graphics-mode       \ default program mode (for testing) 

\ === Coordinate conversion
\ Screen coordinates are Cartesian ( x horizontal, y vertical), with graphics convention (0,0 is top left corner, y is down).  Not using matrix convention ( row i, col j).  
\ Physical Implementation is in linear array, row by row (aka physical coordinates)
\ User coordinates are subset of Screen coordinates, with offsets that prevent user from moving into forbidden areas (menus, etc.)

: xy2lin ( x y -- lin )  \ USER to physical coordinates
\ test: 0 0 xy2lin  = 320, is 3rd row down
    physical-cols *     \ y rows to linear array
    swap 2 *        \ x cols (0..79) to even x-slots (0,2,4,...158) since only evens are visible
    +               \ combine to get linear position
    get-lin-offset + ; \ plus linear offset to reference to user coordinates (0,0) 
: lin2xy ( lin -- x y )  \ physical to USER coordinates
    get-lin-offset -   \ remove offset
    physical-cols /mod ( x y ) 
    swap 2 /          \ to screen char coords (80 per row)
    swap ;
: xy2lin-screen ( x y -- lin )  physical-cols * swap 2 * + ; \ SCREEN to physical coordinates

: init-pen-position ( -- x y ) 0 screen-height 1 - ;
: print-location ( x y -- x y )   2dup ." (x: " swap . ." y: " . ." )" ;    \ TO DO: Add this to status bar

\ === Painting to Screen- Low-level Character Routines
\ color information is encoded into a single byte for use in assembly language routine.  
\ color-code (e.g. $F4) holds background color (bgcolor) in high nybble ($F is white) and foreground (text) color in low nybble ($4 is red)
\ pixel position is in linear encoded form, i.e. row/col are encoded within a single wrap-around array of screen positions

code poke-pixel ( linear-location  color-code  ascii-code  -- )  \ sets a pixel at physical position in linear array with given colors and ascii character.  
    pop cx                  \ load CL with ascii char
    pop ax                  \ load AL with color-code (foreground & background)
    mov ch, al              \ create text pixel (color|char) in (CH|CL)
    pop bx                  \ load BX with screen location
    push ds                 \ preserve data stack
    mov ax, # $B800         \ load AX with graphics address B800 hex (vga memory base address for direct display to screen)
    mov ds, ax              \ set data segment to graphics address (cannot load literal directly, only via register)
    mov 0 [bx], cx          \ display pixel in desired location in graphics memory [B800:bx] (0 data segment offset)
    pop ds                  \ restore data stack
next end-code
\ test: 1000 $FF BL poke-pixel
\ test: 1000 228 42
\ test: 320 228 42 
\ test: 77 20 xy2lin 228 42

code peek-pixel ( linear-location -- color-code ascii-code )
    pop bx              \ load BL with screen location
    push ds             \ preserve data stack
    mov ax, # $B800     \ load AX with graphics address (B800 hex)
    mov ds, ax          \ set data segment to graphics address
    mov cx, 0 [bx]      \ load CX with pixel data from desired location in graphics memory (B800:bx)  (CH|CL)=(color|char)
    pop ds              \ restore data stack
    mov ax, # 0         \ clear AX
    mov al, ch          \ load AX with color-code
    mov ch, # 0         \ clear CH, leaving CL with ascii code
    push ax             \ return color-code
    push cx             \ return ascii code    
next end-code

: set-char ( linloc asc fgcol bgcol -- ) fgbg2col swap poke-pixel ;    \ test: row#3 '*' red yellow
: poke-xy  ( asc fgol bgcol x y -- ) xy2lin 3 roll 3 roll 3 roll set-char ;   \ FIXES UNFORTUNATE STACK DECISION where location is before character and color
: poke-xy-screen  ( asc fgol bgcol x y -- ) xy2lin-screen 3 roll 3 roll 3 roll set-char ;  \ screen to pixel

: get-char ( linloc -- asc fgcol bgcol ) peek-pixel swap col2fgbg ;    \ test: turtle-init 0 20 read-screen-pixel
: read-screen-pixel ( x y -- asc fgcolor bgcolor ) xy2lin get-char ;

: fill-screen ( color start-loc end-loc -- )  \ color (0-15) or color-name ( black, white, yellow, red, ... )
    rot dup fgbg2col -rot                     \ create color code 
    BL -rot                                   \ fill with blanks    
    over - 2/ 1 +                             \ length of physical array to fill but only the evens (odd cells go green in this video mode) 
( colorfgbg BL start-loc length )             
    0 do ( colorfgbg BL start-loc )           \ Only start-loc needs to change, rest are constant.  Looped instructions: 7.
        >r 2dup r@ -rot poke-pixel r> 2 +     \ increment by 2 as only even cells are visible
    loop 3drop ;  

: black-screen  ( -- ) black physical-zero physical-end fill-screen ;

: hline-draw ( color x1 x2 y -- )
    -rot swap do ( color y ) 
        i over xy2lin-screen bl 3 pick dup set-char
    loop 2drop ;
: vline-draw ( color x y1 y2 -- )
    swap do ( color x )
        dup i xy2lin-screen bl 3 pick dup set-char
    loop 2drop ;
: borders-draw ( -- )  \ all coords are in physical space
    get-border-color  0 80  1  hline-draw
    get-border-color  0 80 23  hline-draw
    get-border-color  0  1 23  vline-draw
    get-border-color 79  1 23  vline-draw ;


\ === PIXELS & PICTURE
\ A pixel is a 3 byte triple of ascii code, foreground color, background color, stored in that order.
\ Picture is a single array (nrow x ncol x 3 bytes) of pixels used to save/restore picture to/from memory.
\ Pixel variable is a temporary storage to hold 1 pixel information while user is deciding, in case the picture gets impacted (e.g. home, penup/down, etc.)
3   constant   pixel-sz
create         pixel          pixel-sz   bytes allot    \ variable - temp storage for 1 pixel
screen-cols screen-rows *     constant   num-pix-scr
num-pix-scr    pixel-sz *     constant   pic-sz               
create  picture   pic-sz bytes allot                    \ array - memory block to hold full screen
: picture-clear ( --- )  picture pic-sz 0 fill ;        \ zero out memory

: pix!+ ( asc fgcolor bgcolor pix-ptr -- pix-ptr' ) >r rot r> c!+ rot swap c!+ c!+ ;    \ store pixel, preserving pixel order
: pix@+ ( pix-ptr -- asc fgcolor bgcolor pixl-ptr' ) c@+ swap c@+ swap c@+ swap ;       \ fetch pixel
: set-pixel+ ( 1 2 3 -- ptr' ) pixel pix!+ ; 
: get-pixel+ ( -- 1 2 3 ptr' ) pixel pix@+ ; 
: cache-pixel ( x y -- x y ) 2dup read-screen-pixel  set-pixel+  drop  ;

\ Changing a pixel
\ When turtle is on the pixel (show-heading), where turtle has moved on from with pendown (paint-xy) or with penup (restore-pixel), where backspace (erase-xy) 
: show-heading  ( x y h# -- )   -rot xy2lin swap get-heading-symbol get-pen-color get-cursor-color set-char ;  \ draw symbol on screen 
: paint-xy      ( x y -- x y )  2dup xy2lin BL get-pen-color dup ( loc asc fg bg ) set-char ; \ test 3 12
: restore-pixel ( x y -- x y )  2dup xy2lin get-pixel+ drop set-char ;
: erase-xy      ( x y -- x y )  2dup xy2lin BL white white set-char ;

\ Changing a picture: note each pixel is a byte in memory, but occupies two bytes on screen.
: pic-scr2mem ( -- ) picture num-pix-scr 0 do i 2 * ( pic linloc ) swap >r get-char r> pix!+ loop drop ;   \ save picture from screen to memory
: pic-mem2scr ( -- ) picture num-pix-scr 0 do i 2 * swap pix@+ >r set-char r> loop drop ;     \ load memory from memory to screen

\ === Paint to Menu
\ !! HOW IT SHOULD WORK (how it works seems more complicated)
\ The menu display system should be simple: draw the menu, refresh dynamic content when it changes.
\ Do this to a virtual copy of the menu so that it persists even with screen changes (e.g. display help system, clear screen, etc.)
\ Text should use the default fgbg unless in a special mode, in which case this is provided instead.
\ !! END CODE REVIEW

\ Menu Features - screen coordinates
0   constant menu2-h#      \ heading (x=.)
13  constant menu2-c#      \ color (x=.)
24  constant menu2-r#      \ recording (x=.)
0   constant menu1-row     \ top menu (y=.)
24  constant menu2-row     \ bottom menu (y=.)

: menu1-text ( -- )  " Commands: [ARROWS]Move/Rotate [BKSPC]Erase [']Penup/Dn [H]ome [Esc]Clear [?]HELP" ;  \ 80 char max
: menu2-text ( -- )  " -- [X/C]olor:------ [T]:------------ [N]extSkill [1-9]Play [S/U]Save/Undo [Q]uit" ;  \ 80 char max   \ [L]oop:--

\ TESTER CODE FOR BETTER WRITING OF MENUS
( linloc c-adr nchar fg bg -- )

\ writing characters and strings
: write-char+ ( col L ca  -- col L' ca' ) 
        over 2 + -rot ( col l' l c )            \ save next location (2 bytes ahead - see graphics explanation: 80 screen positions, 160 physical positions)
        dup 1 + -rot c@ ( col l' c' l char )    \ get char and save next pointer position
        4 pick swap ( col l' c' l col char )    \ arrange parameter order
        poke-pixel ;
: write-string ( L ca n fg bg -- ) fgbg2col swap >r -rot r> ( col L ca n ) 0 do write-char+ loop ;
: test-setup 0 15 xy2lin-screen menu1-text red lime ;  \ the 0 15 are just the menu row and col (screen coords) for starting the string write
: test test-setup write-string ;
\ CONSIDER changing the parameter order to set-char

\ menu parameters: row location (y), foreground/background color for text, and col position on menu to update dynamic information (in screen coordinates)
: menu1-row 0 ;
: menu1-fgbg  white navy ;
: menu2-row 24 ;
: menu2-fgbg  white navy ;
: menu2-heading-position ( -- loc )  menu2-h# menu2-row  ;
: menu2-color-position ( -- loc )    menu2-c# menu2-row  ;  
: menu2-rec-pos ( -- loc) menu2-r# menu2-row ;
: create-rec-fgbg      \ based on rec state
    create 
        white c, slate c,     \ if 0 
        red c, yellow c,      \ if 1
        navy c, yellow c,     \ if 2
        white c, red c,       \ if 3
        white c, orange c,    \ if 4
    does> ( -- )
       get-rec-state 2 * bytes + c@+ swap c@ ; 
create-rec-fgbg menu2-rec-fgbg
: menu2-attn-fgbg ( -- ) red lime ;  \ color scheme to highlight text

\ creating menus
: new-menu 
    create ( menu-text_c-addr|nchar  menu-row  fg bg  | <name> -- )   \ create dictionary entry
    fgbg2col c,                                                       \ create default menu color code
    physical-cols * ,                                                 \ physical offset to start on screen at (0,menu-row)
    here dup , screen-cols 2dup dup bytes allot BL fill drop swap cmove   \ allocate memory for string, initialized with blanks, then copied with menu-text
does>  ( -- )                                                          \  places address of entry on stack 
    c@+ swap  \ fgbg
    @+ swap   \ start-loc physical   
( col L ca )
    screen-cols 0 do write-char+ loop                                  \ print menu to screen
    3drop ;

menu1-text  menu1-row  menu1-fgbg   new-menu  menu1
menu2-text  menu2-row  menu2-fgbg   new-menu  menu2

: test-pixel  1000 31 'X' poke-pixel ;
: setup-menutest menu2-text menu2-row menu2-fgbg ;
: test-create ( <menu3> ) setup-menutest new-menu ;

\ dynamically updating menus
: menu-write ( c-addr length fg bg xcol-screen yrow-screen -- )
    xy2lin-screen             \ calc physical offset
    -rot >r >r -rot r> r>     \ prepare stack
    ( linloc c-addr nchar fg bg )
    write-string 3drop ;
: test-menu-write  5 color-names red yellow 20 23 menu-write ;

: menu2-show-heading ( -- )  get-heading  get-heading-string  2  menu2-fgbg menu2-h# menu2-row menu-write ;
: get-pencolor-name ( -- index ) get-penup-state if penup-col-index else get-pen-color then color-names ;
: menu2-show-color ( -- )  get-pencolor-name get-penup-state if menu2-attn-fgbg else menu2-fgbg then menu2-c# menu2-row menu-write ;
: menu2-show-recording-state ( -- )  get-rec-str  menu2-rec-fgbg  menu2-r# menu2-row menu-write ;

: menu2-refresh ( -- ) menu2-show-heading  menu2-show-color  menu2-show-recording-state ;  \ paints the dynamic portions
: paint-menus  ( -- ) menu1 menu2 ;  \ paints static menus
: paint-screen ( color -- ) physical-zero physical-end fill-screen paint-menus borders-draw ;

\ TESTER CODE
: menu2-screen-pos ( x y -- ... ) ;

\ === Show/Hide Turtle
create visible-state 1 cells allot  \ VARIABLE - BOOLEAN - turtle visibility state -- true if visible, false if hidden
: set-visible ( -- ) true visible-state ! ;
: set-hidden  ( -- ) false visible-state ! ;
: is-visible? ( -- t/f) visible-state @ ;
: turtle-hide ( x y -- x y )  restore-pixel ;   \ erase current location
: turtle-show ( x y -- x y )  is-visible? if 2dup get-heading show-heading else turtle-hide then ;

\ === Movement Control
\ movement is either forward or backwards along the axis of the current heading, with screen wrap-around
\ moving forward paints, moving backward erases
\ coordinates (x,y) are semi-cartesian: over and down.  (0,0) is the top left of the user space, i.e. excluding the screen border and menus.
\ movement = 1) update coordinates to a new pixel, 2) paint or restore previous pixel, 3) new pixel is cached, 4) show turtle in new pixel.

\ screen wrap-around
: wrap-y ( y -- y* ) screen-height mod ;
: wrap-x ( x -- x* ) screen-width mod ;

\ movement along direction axes
: move-N-fwd   ( x y -- x' y' ) 1 - wrap-y ;
: move-N-back  ( x y -- x' y' ) 1 + wrap-y ;
: move-E-fwd   ( x y -- x' y' ) swap 1 + wrap-x swap ;
: move-E-back  ( x y -- x' y' ) swap 1 - wrap-x swap ;
: move-S-fwd   ( x y -- x' y' ) move-N-back ;
: move-S-back  ( x y -- x' y' ) move-N-fwd ;
: move-W-fwd   ( x y -- x' y' ) move-E-back ;
: move-W-back  ( x y -- x' y' ) move-E-fwd ;
: move-NE-fwd  ( x y -- x' y' ) move-N-fwd move-E-fwd ;
: move-NE-back ( x y -- x' y' ) move-N-back move-E-back ;
: move-SW-fwd  ( x y -- x' y' ) move-NE-back ;
: move-SW-back ( x y -- x' y' ) move-NE-fwd ;
: move-NW-fwd  ( x y -- x' y' ) move-N-fwd move-W-fwd ;
: move-NW-back ( x y -- x' y' ) move-N-back move-W-back ;
: move-SE-fwd  ( x y -- x' y' ) move-NW-back ;
: move-SE-back ( x y -- x' y' ) move-NW-fwd ;

\ create motion array of movements in h-index (rotation cycle) order
: create-fwd-motions
create ( | <name> -- )     
    ['] move-N-fwd   , 
    ['] move-NE-fwd  ,  
    ['] move-E-fwd   , 
    ['] move-SE-fwd  ,  
    ['] move-S-fwd   , 
    ['] move-SW-fwd  ,  
    ['] move-W-fwd   , 
    ['] move-NW-fwd  , 
does> ( x y -- x' y' ) get-heading cells + @ execute ;
: create-back-motions
create ( | <name> -- )     
    ['] move-N-back   , 
    ['] move-NE-back  ,  
    ['] move-E-back   , 
    ['] move-SE-back  ,  
    ['] move-S-back   , 
    ['] move-SW-back  ,  
    ['] move-W-back   , 
    ['] move-NW-back  , 
does> ( x y -- x' y' ) get-heading cells + @ execute ;
CREATE-FWD-MOTIONS    move-fwd-headings   ( x y -- x' y' )
CREATE-BACK-MOTIONS   move-back-headings  ( x y -- x' y' )

: move-fwd ( x y -- x' y' )   get-penup-state if restore-pixel else paint-xy then  move-fwd-headings  ;
: move-back ( x y -- x' y' )  get-penup-state if restore-pixel else erase-xy then move-back-headings  ;  \ erases only in pendown mode
                                 
\ === Key logger / Diary function
\ For nested macros to work, we need a diary stack.

9 constant num-diaries                  \ number of distinct recordable macros
1000 bytes constant diary-length        \ max keypress codes in each macro
create   diary   num-diaries diary-length *  allot      \ allocate diary memory
diary num-diaries diary-length * 0 fill                 \ zero out diary memory

create d# 1 bytes allot   \ VARIABLE - holds active diary number - digits 0 to 8
: switch-to-diary ( d# -- ) dup d# c! update-diary-string ;
: get-diary-num ( -- d# ) d# c@ ;
: next-diary ( -- ) get-diary-num 1 + num-diaries mod  ( d# ) switch-to-diary ;

create cached-d# 1 bytes allot  \ VARIABLE - holds original diary number so that nested macro playbacks end by showing the original diary number.
: cache-diary-num ( d# -- ) cached-d# c! ;
: restore-diary-num ( -- ) cached-d# c@ switch-to-diary ;

create diary-end num-diaries cells allot                       \ ARRAY - holds diary ends 
diary-end num-diaries cells 0 fill      \ zero out mem
: set-diary-end ( d# dptr -- ) swap cells diary-end + ! ;  \ store position 
: get-diary-end ( d# -- dentry ) cells diary-end + @ ;
: diary-full?  ( d# dptr -- t/f ) 2dup swap get-diary-end > ;

create diary-pos num-diaries cells allot  \ ARRAY - holds next free place in each diary, which is the stop point for playback .
diary-pos num-diaries cells 0 fill      \ zero out mem
: set-diary-pos ( d# dptr -- )  swap cells diary-pos + ! ;         \ store position
: get-diary-pos ( d# -- dptr )  cells diary-pos + @ ;
: diary-start ( d# -- d# dptr )  dup diary diary-length rot * + ;  \ returns start position
: init-diaries ( -- )
    0 diary-start 
    num-diaries 0 do
       2dup set-diary-pos               \ set start
       diary-length + 2dup 1 - set-diary-end  \ set end
       ( d# dptr ) swap 1 + swap 
    loop 2drop ;
: rewind-diary-pos ( d# -- ) get-diary-pos 1 - set-diary-pos ;  \ removes last entry

\ ==Viewing the Diary==
: show-row ( dptr -- dptr' ) cr 10 0 do space c@+ emit loop space space
                  10 0 do space c@+ emit loop 
           pq ; 
: cue-to-start ( d# -- dptr ) diary diary-length rot * +  ;  \ cue to diary start position
: fetch-end    ( d# -- dptr ) cells diary-pos + @ ; 
: dump-diary ( diarynum -- )  \ show entire diary, filled or not
    1 - ( d# ) cue-to-start 
    diary-length 20 / 0 do show-row loop drop ;    
: see-diary ( diarynum -- ) 
    1 - ( d# ) dup cue-to-start swap fetch-end over - 20 / 1 + 0 do show-row loop drop ;

create diary-recording 1 bytes allot   \ VARIABLE - true if recording
: is-recording? ( -- t/f ) diary-recording c@ ;
: set-diary-recording ( -- ) true diary-recording c! ;
: clear-diary-recording ( -- ) false diary-recording c! ;

: show-recording ( -- )  set-diary-recording  rec-state-recording  set-rec-state  menu2-show-recording-state ;
: flash-recording ( -- )  rec-state-reminder  set-rec-state  menu2-show-recording-state 300 ms show-recording ;

: is-paused? ( -- t/f ) get-rec-state rec-state-paused = ;
: start-recording ( -- ) get-diary-num  diary-start set-diary-pos show-recording  ;     \ record macro from beginning
: resume-recording ( -- ) show-recording ;        \ record macro from previous position
: stop-recording ( -- )  clear-diary-recording     rec-state-ready  set-rec-state    menu2-show-recording-state ;
: pause-recording ( -- )  clear-diary-recording     rec-state-paused set-rec-state  menu2-show-recording-state ; 

: warn-diary-full ( -- )  rec-state-full set-rec-state menu2-show-recording-state 700 ms stop-recording ;
: diary-write-entry ( d# asc -- )  over get-diary-pos c!+ ( d# dptr' ) 
        diary-full? if 2drop warn-diary-full else set-diary-pos then ;




\ ===KEY PROCESSING ENGINE===
\ We find what action turtle should take based on user keypress using a key mapping lookup table which associates every ascii key with the executable token (xt) giving its intended action (if it has one).  Non mapped keys are ignored.
\ The actual lookup table is below, after the actions have all been defined (search =keymap=)
create   keymap   256 cells allot                  \ creates storage for 256 executable tokens (xt) logged at their ascii code offset location
: add-key ( xt keycode -- ) cells keymap + ! ;     \ stores xt into keymap table at keycode offset location
: ignore-input  ( asc -- ) drop ;
: init-keymap ( -- ) ['] ignore-input 255 0 do dup i add-key loop drop ;  \ initialize every key in keymap to ignore input; 
: create-mapping  ( keymap <name> -- )   \ create-mapping overwrites mapped keys
create ,                \ compiles keymap base address into <named> instruction
does> ( x y asc -- x y )  \ key processing engine: defines what the created instruction (e.g. turtle-do) will do when called
    @ ( keymap)  over cells + @ ( x y asc xt ) execute ( x y )  \ all key commands consume asc and restore rest of stack 
    turtle-show menu2-refresh ;
keymap create-mapping TURTLE-DO

\ ??? TESTER CODE
create diary-stack 16 bytes allot  \ STACK - saves context for nested macros
create tods 1 cells allot  \ VARIABLE - holds top of diary stack
: init-diary-stack diary-stack 16 $FF fill  \ set memory to $FF
    diary-stack tods ! ;   \ set top of stack
init-diary-stack

: set-tods ( dstack-ptr --  ) tods ! ; 
: get-tods ( -- dstack-ptr  ) tods @ ; 
: push-diary ( d# -- ) get-tods c!+ set-tods ;
: pop-diary ( -- d# ) tods @ 1 - dup c@ swap set-tods ;
: dump-diary-stack ( -- ) diary-stack 16 0 do c@+ . loop drop ;
: see-diary-stack ( -- ) tods @ 1 + diary-stack do i c@ . loop ;
: test-diarystack ( 3 2 1 0 -- )   \ run it on command line
    dump-diary-stack see-diary-stack 
    3 2 1 0 push-diary push-diary push-diary push-diary 
    dump-diary see-diary 
    pop-diary . pop-diary . pop-diary . ;


: diary2d# ( asc - d# ) a2d 1 - ;
: macro-playback ( x y asc -- x' y' )
    diary2d#  dup switch-to-diary  diary-start
    swap get-diary-pos swap
    2dup <= if 2drop
    else 
        do i c@ turtle-do loop 
    then ;


\ ============ Control keys
\ Keyboard commands are associated with their ASCII codes.
\ Design is child friendly, with upper and lower case letters mapped to the same command so the program works even if caps lock gets pressed.
\    / is mapped to help (?) since ? requires use of SHIFT which is not easy for a child
\    (numbers are not affected by Caps Lock)
$C8          constant    fwd-key            \ 200d
$CD          constant    right-key          \ 205d
right-key    constant    cw-key
$CB          constant    left-key           \ 203d
left-key     constant    ccw-key
$8           constant    backspace
backspace    constant    erase-key
$27          constant    penupdown-key      \ tick
'c'          constant    color-fwd-key      \ change color
'C'          constant    color-fwd-key2     \ change color
'x'          constant    color-back-key     \ change color
'X'          constant    color-back-key2    \ change color
'n'          constant    diary-key          \ RENAMED NEXT
'N'          constant    diary-key2         \ RENAMED NEXT
't'          constant    record-key         \ RENAMED TEACH
'T'          constant    record-key2        \ RENAMED TEACH
'1'          constant    play-macro-1-key
'2'          constant    play-2-key
'3'          constant    play-3-key
'4'          constant    play-4-key
'5'          constant    play-5-key
'6'          constant    play-6-key
'7'          constant    play-7-key
'8'          constant    play-8-key
'9'          constant    play-9-key
's'          constant    save-key
'S'          constant    save-key2
'u'          constant    undo-key
'U'          constant    undo-key2
';'          constant    show-hide-key      \ turtle hide/view 
'h'          constant    home-key 
'H'          constant    home-key2
esc-key      constant    clear-screen-key
'?'          constant    help-key
'/'          constant    help-key2           \ Usability: in case a child does not know how to use shift key
'q'          constant    quit-key
'Q'          constant    quit-key2

\ === TURTLE PROCESSING COMMANDS
\ These commands are ALWAYS logged if recording is active.
: log-key-if-recording ( asc -- asc ) is-recording? if get-diary-num over ( asc d# asc ) diary-write-entry then ;
: turtle-move ( x y asc -- x' y' ) log-key-if-recording  fwd-key = if move-fwd else move-back then cache-pixel set-visible ;
: adjust-heading ( h-delta -- )  get-heading + 8 mod set-heading ;  \ each h-delta nudges cw by 45 degrees, negative ccw by 45 deg
: heading-change ( asc -- )
    log-key-if-recording ( asc -- )
    cw-key = if 1 else -1 then adjust-heading
    set-visible ;
: do-penup   ( -- )  true set-penup-state save-prev-pen-color set-penup-color ; \ pen-up mode has custom appearance
: do-pendown ( -- )  false set-penup-state  restore-prev-pen-color  ;
: penupdown ( x y asc -- x y ) log-key-if-recording  drop get-penup-state if do-pendown else do-penup then ;
\ ??? CODE REVIEW -- names should be swapped for next two words...
: turtle-color ( asc -- ) color-fwd-key = if 1 else -1 then get-pen-color  +  16  mod  set-pen-color ;  \ +1/-1 cycle fwd/bk to next color of 16
: color-change ( x y asc -- x y ) log-key-if-recording   get-penup-state if do-pendown then turtle-color ; 

\ These commands are SOMETIMES logged if recording is active.
: log-key-if-not-current-diary ( macro-asc -- macro-asc ) dup diary2d# get-diary-num = ( macro-asc t/f ) 
    if flash-recording else get-diary-num over diary-write-entry then ;

: play-macro ( x y asc -- x y )
    is-recording?  
if  
    log-key-if-not-current-diary    \ nested macros allowed (mixture of primary and derived commands) but not recursion (calling oneself)
    get-diary-num  push-diary       \ diary stack
    pause-recording                 \ don't want the individual keypresses to be recorded, just the macro-name
    macro-playback                  \ want to see the macro playing on the screen as constructing the new macro
    pop-diary switch-to-diary       \ store diary context from the stack
    resume-recording                \ allow recording to continue from same position
else 
    macro-playback                  \ replay macro
then ;


\ These commands are not logged.
: save-point ( x y asc -- x y ) drop is-recording? if flash-recording then pic-scr2mem ;   \ save a restore point
: restore-point ( x y asc -- x y ) drop is-recording? if flash-recording else pic-mem2scr then ;  \ restore to last saved image (saving before a macro run)

: turtle-showhide ( x y asc -- x y ) drop is-recording? 
    if flash-recording 
    else is-visible? if set-hidden else set-visible then 
    then ; \ toggle turtle visibility
: turtle-home ( -- )
    get-penup-state dup >r                         \ store penup-state
    if else do-penup  then turtle-hide             \ don't leave mark in last position 
    2drop init-pen-position init-heading 
    cache-pixel                               \ don't clobber home position
    r> if else do-pendown then ;                   \ usability: restore the pen to the user's last state
: go-home ( x y asc -- x y ) drop is-recording? if stop-recording then turtle-home ;    \ reset turtle to home position

: diary-next ( x y asc -- x y ) drop is-recording? if stop-recording then next-diary ;
: recording-onoff ( x y asc -- x y )  drop is-recording? if stop-recording else start-recording then ;

: safe-cr ( -- ) pic-scr2mem cr pic-mem2scr ;     \ buffering and redrawing screen to avoid damage from carriage return
: confirm-quit ( -- t/f )
    safe-cr  quit? if true else cr pic-mem2scr false then ;
: is-quit? ( asc -- asc t/f ) 
        quit-key over <> over quit-key2 <> and  ( asc t/f )  \ checking both cases (q and Q)
        if false                                             \ no quit request, continue
        else                                                 \ quit request
           is-recording? 
                if flash-recording false                     \ not allowed to quit while recording
                else confirm-quit if true else false then
                then
        then ;
: clear-screen ( x y asc -- x y ) drop  is-recording? if stop-recording then
        safe-cr  clear?   cr                                                \ save screen in case user changes mind
        if    cr  turtle-home  get-screen-color  paint-screen  set-visible 
        else  cr pic-mem2scr                                                \ restore saved screen
        then  ;
        
: show-help ( -- )  pic-scr2mem  black-screen  welcome pause-key  pic-mem2scr ;
: help-screen ( x y asc -- x y ) drop show-help ;

: map-key-commands ( -- )   \ actions saved as executable tokens into =keymap= at their ascii keycode offset
    init-keymap
\         xt               keycode
    ['] turtle-move       fwd-key              add-key
    ['] turtle-move       erase-key            add-key
    ['] heading-change    left-key             add-key
    ['] heading-change    right-key            add-key
    ['] penupdown         penupdown-key        add-key
    ['] color-change      color-fwd-key        add-key
    ['] color-change      color-fwd-key2       add-key
    ['] color-change      color-back-key       add-key
    ['] color-change      color-back-key2      add-key
    ['] turtle-showhide   show-hide-key        add-key
    ['] go-home           home-key             add-key
    ['] go-home           home-key2            add-key
    ['] diary-next        diary-key            add-key
    ['] diary-next        diary-key2           add-key
    ['] recording-onoff   record-key           add-key
    ['] recording-onoff   record-key2          add-key
    ['] play-macro        play-macro-1-key     add-key
    ['] play-macro        play-2-key           add-key
    ['] play-macro        play-3-key           add-key
    ['] play-macro        play-4-key           add-key
    ['] play-macro        play-5-key           add-key
    ['] play-macro        play-6-key           add-key
    ['] play-macro        play-7-key           add-key
    ['] play-macro        play-8-key           add-key
    ['] play-macro        play-9-key           add-key
    ['] save-point        save-key             add-key
    ['] save-point        save-key2            add-key
    ['] restore-point     undo-key             add-key
    ['] restore-point     undo-key2            add-key
    ['] clear-screen      clear-screen-key     add-key
    ['] help-screen       help-key             add-key
    ['] help-screen       help-key2            add-key
;

\ === MAIN PROGRAM
: turtle-init ( -- x y )  \ Initialization code
    set-video-mode   set-lin-offset   picture-clear   init-diaries
    attach-rec-msgs   stop-recording  map-key-commands
    red set-pen-color   white  set-screen-color    silver set-border-color 
    cr get-screen-color paint-screen   
    set-visible   false set-penup-state   0 switch-to-diary   init-heading   init-pen-position ( -- x y )   menu2-refresh  
    cache-pixel  turtle-show ;

: q ( -- ) \ quit
    turtle-hide is-recording? if stop-recording then
        cr ." To restart, type TURTLE-LOGO"
        cr ." To return to Windows, type EXIT." 
        bye  \ exits Forth to DOS.  Comment out to stay within Forth e.g. during development
        ;  

: turtle-resume ( x y -- )
    pic-mem2scr
    begin                  \ input processing loop
        key ( x y asc ) is-quit? if drop true else turtle-do false then  
    until ( x y ) q ;    

: turtle-go ( -- x y )  
    show-help  
    turtle-init  ( x y ) 
    begin                  \ input processing loop
        key ( x y asc ) is-quit? if drop true else turtle-do false then  
    until ( x y ) q ;
: turtle-logo ( -- ) turtle-go ;   \ alternative

\ ================================================ END OF MAIN PROGRAM



\ ================================================ TESTING - SUPPLEMENTARY CODES
\ === instructions
: kp ( x y asc -- x y ) turtle-do  ; 
: .. ( -- ) pause-quit ;
: pen-adjust ( -- ) penupdown-key kp ;
: recording-adjust ( -- ) record-key kp ;
: diary-adjust ( -- ) diary-key kp ;
: turtle-adjust ( -- ) show-hide-key kp ;
: test-home ( -- ) home-key kp ;
: next-color ( -- ) color-fwd-key kp ;
: prev-color ( -- ) color-back-key kp ;

\ === macros
: left-45 ( -- ) left-key kp ;
: left-90 ( -- ) 2 0 do left-45 loop ;
: left-180 ( -- ) 4 0 do left-45 loop ;
: right-45 ( -- ) right-key kp ;
: right-90 ( -- ) 2 0 do right-45 loop ;
: right-180 ( -- ) 4 0 do right-45 loop ;
: fwd-n ( n -- ) 0 do fwd-key kp loop ;
: erase-n ( n -- ) 0 do erase-key kp loop ;

: draw-half-square ( -- )  2 fwd-n   right-90  5 fwd-n  right-90 ;
: house-square ( -- ) draw-half-square draw-half-square ;
: house-roof ( -- ) 3 fwd-n right-45 2 fwd-n right-45 1 fwd-n right-45 2 fwd-n right-45 1 fwd-n ;
: house-complete ( -- ) pen-adjust 2 fwd-n left-90 3 fwd-n left-90 pen-adjust ;
: draw-house ( -- ) house-square house-roof house-complete ;
: draw-spiral ( -- ) 
    4 fwd-n   right-90  8 fwd-n  right-90
    3 fwd-n   right-90  1 fwd-n  color-fwd-key kp  5 fwd-n  right-90 ;

\ === test harnesses
: test-show-color-names ( -- ) 17 0 do i color-names cr type loop ;

: test-poke-pixel ( -- )
    row#3 $5F 'S' poke-pixel  \ interactive visible start point
    physical-end $8F 'E' poke-pixel      \ end
    3 1 xy2lin $A0 'A' poke-pixel   
    68 12 xy2lin $C0 'B' poke-pixel
    69 13 xy2lin $40 'C' poke-pixel ;

: test-probe-diary  ( -- )
    0 switch-to-diary
    init-diaries  set-diary-end
    ." diary # " get-diary-num . 
    ." diary-pos: "  get-diary-pos u. 
    ." diary-end: "  get-diary-end u. 
    cr ;

: test-diaries ( -- )  
    1 fwd-n '1' kp ..   \ checking that no playbacks occur, no errors occur, but diary changes with each keypress
    1 fwd-n '2' kp ..
    1 fwd-n '3' kp ..
    1 fwd-n '4' kp ..
    1 fwd-n '5' kp ..
    1 fwd-n '6' kp ..
    1 fwd-n '7' kp ..
    1 fwd-n '8' kp ..
    1 fwd-n '9' kp ..
    right-90
    1 fwd-n '1' kp ..    ;
    
\ TO TEST fullness for diaries, set diaries from 1000 to 15.  Remember to set the diary length back to 1000 chars.
: test-diary-fullness-1-channel ( -- )  \ check that all 9 diaries work correctly.
    record-key kp ..
    8 fwd-n right-90 ..  \ 10 moves
    1 fwd-n right-90 next-color ..  \ 14 moves 
    right-45 .. \ 15th move triggers fullness, does not record, but executes
    left-45 ..  ; \ corrects  
    
: test-diary-fullness ( -- )  \ recording, fullness, and playback.
    turtle-init
    test-diaries
    test-diary-fullness-1-channel diary-key kp ..
    test-diary-fullness-1-channel diary-key kp ..
    test-diary-fullness-1-channel diary-key kp ..
    test-diary-fullness-1-channel diary-key kp ..
    test-diary-fullness-1-channel diary-key kp ..
    test-diary-fullness-1-channel diary-key kp ..
    test-diary-fullness-1-channel diary-key kp ..
    test-diary-fullness-1-channel diary-key kp ..
    test-diary-fullness-1-channel diary-key kp ..
    '1' kp ..
    '2' kp ..
    '3' kp ..
    '4' kp ..
    '5' kp ..
    '6' kp ..
    '7' kp ..
    '8' kp ..
    '9' kp .. ;    

: test-multi-diary-recording ( -- )  \ draws a house all together, then in recordable bits, then by playing the recorded bits, then stitching the recorded bits into one recording
    turtle-init
    draw-house ..
    recording-adjust house-square .. recording-adjust .. diary-adjust ..
    recording-adjust house-roof .. recording-adjust .. diary-adjust ..
    recording-adjust house-complete .. recording-adjust .. diary-adjust ..
\    '1' kp ..
\    '2' kp ..
\    '3' kp ..
    recording-adjust '1' kp .. '2' kp .. '3' kp .. recording-adjust .. diary-adjust ..
    '4' kp .. ;    

    
: test-erase ( -- )
    turtle-init
    right-45    3 fwd-n   3 0 do next-color loop   3 fwd-n   3 0 do next-color loop   3 fwd-n  ..
    right-180 ..
    pen-adjust ..
    7 fwd-n ..
    5 erase-n ..  \ bug!
    right-90 .. 
    1 erase-n .. 
    right-90 .. 
    1 fwd-n .. ; \ bug!

: test-home-key ( -- )
    turtle-init
    draw-spiral draw-spiral draw-spiral ..
    test-home ..
    6 fwd-n pen-adjust ..
    draw-spiral ..
    test-home ..
    right-45 6 fwd-n ..
    test-home ..
    right-45 8 fwd-n ..
    pen-adjust ..
    test-home ..
    right-45 14 fwd-n ..
    test-home ..
    right-45 14 fwd-n pen-adjust ..
    test-home .. ;

: turtle-test ( -- x y ) 
    turtle-init ( -- x y )
\ testing fwd,right,left
    fwd-key kp right-45  2 fwd-n left-90 2 fwd-n ..   
\ testing erase
    right-45 2 fwd-n .. 
    2 erase-n ..  
\ testing colorfwd/colorback
    3 0 do next-color loop 2 fwd-n ..
    2 0 do prev-color loop 2 fwd-n ..
\ testing penup/pendown
    right-90 8 fwd-n ..
    left-180 pen-adjust .. 
    4 fwd-n ..
    pen-adjust ..
    2 0 do next-color loop ..
    2 fwd-n ..
\ testing home
    test-home ..
    left-90 4 fwd-n ..
\ testing macro-key before any record
    5 0 do next-color loop .. 
    play-macro-1-key kp ..  \ nothing bad should happen
    next-color ..
\ testing recording and pressing macro-key while recording
    recording-adjust ..
    play-macro-1-key kp ..  \ recording notification should flash
    next-color .. 
    1 fwd-n .. 1 fwd-n .. 1 fwd-n .. right-90 .. \ one leg of square
    recording-adjust ..
\ testing macro replay
    play-macro-1-key kp ..
    play-macro-1-key kp ..
    play-macro-1-key kp ..
\ testing clear-screen
    clear-screen-key kp  .. \ press yes
    left-180
    play-macro-1-key kp ..
    play-macro-1-key kp ..
    play-macro-1-key kp ..    
    clear-screen-key kp  .. \ press no
\ testing help
    left-90 left-key kp
    play-macro-1-key kp ..
    play-macro-1-key kp ..
    play-macro-1-key kp ..
    play-macro-1-key kp ..
    help-key kp ..
\ testing hide and the various keys that force visibility
    left-90 7 fwd-n ..
    turtle-adjust ..
    next-color ..
    turtle-adjust ..
    1 fwd-n ..
    turtle-adjust ..    
    right-45 .. 
    turtle-adjust .. 
    test-home ..
    turtle-adjust ..
\ test quit
    quit-key kp ;

: fill-screen-test ( -- )
  16 0 do
    i physical-zero physical-end fill-screen pause-quit
  loop ;
4 constant thickness
: color-bg-loc ( lin -- color )  \ pick background color based on location
    lin2xy drop ( j )
    thickness / 16 /mod drop ( color )  ; \ cycle through 16 colors
: color-demo ( -- )  
    set-video-mode
    get-lin-offset
    begin ( loc )
        dup physical-end < 
    while ( loc )
        dup BL over color-bg-loc dup set-char
        2 +  \ only even cells are visible in this graphics mode.  Odd cells go green
    repeat
    drop
    pause-quit ;

: fgbg-demo ( -- )
    set-video-mode
    get-lin-offset
    begin ( loc )
        dup physical-end <
    while ( loc)
        dup 0 get-heading-symbol over color-bg-loc dup set-optim-bg-color get-cursor-color set-char
        2 + 
    repeat
    drop
    pause-quit ;

\ =================================================================
\ TODO:
\  2018-08-20--Mon--23:56 - Restarted

\ Usability - Make the SKILL number highlighted and long with the [N] in new skill  OR have it [S]kill #X
\ Feature: Keyboard shortcut list (K mode) -- pulls out a side panel that then goes away.
\ Bug: how to prevent accidentally pressing record and overwriting previously recorded macro?  Prompt with "Are you sure?"  --- That skill has already been used, are you sure?  And pull out the skill code...
\ Bug: Add second class macros (macros that call themselves) -- why does this not work?
    \ BUG: Only 1 level recording nesting is allowed?  So can't do (1,1), (2,2,2) and combine into (1,1,2,2,2)? (Yahtzee feature)
    \ Feature: Allow higher level nesting in which macro calls other macro (but ok still not to have recursion, i.e. macro calling itself which is expressly forbidden)
\ DONE: allow nested recording of macros (i.e. record macro numbers so macros can be nested within other macros and all played back -- BUGGY - need to maintain a recursion stack for reading the macros back )
\ FIXED: cannot do recursion (i.e. cannot record macro calling itself)
    \ Feature: Consider: support recursive macros (calling themselves?)
\ Feature: Add L for loop -- and you enter how many times to issue that instruction.  (L mode).  Can do it stepwise (with press any key).  This is the start of robotic navigation programming.
\ Feature: Add coordinates to status bar (useful for geometry).  Turtle is at 0,0, and he moves into the 1st quadrant.  (G mode) (Geometry mode)
    \ Feature: display numbered grid (numbered rows and columns according to elementary math convention: H is the origin (0,0) then positive to right and up)
    \ Feature: show coordinate location (trouble shoot with .s on status bar)
\ Feature: 'x' to toggle to text mode, to allow adding text in direction of cursor.  Possibly an auto-date feature when saving the picture.
\ Feature: Have turtle prompt children for certain things.
\ Feature: built in exercises -- shows screen, then asks child to replicate it...
\ Feature: record all key presses, and auto-save, so can quit, restore, save, undo full stack, re-do, and reproduce anything
    \ Feature: save macros to file for reloading later (in case program quits).  Allows user to load previous creations, e.g. trees, flowers, birds, sun, house.  
\ Feature: Save picture, screenshot, or generate print-out
    \ Feature: save screen to disk, load from disk
    \ Feature: allow full-screen view - without menus (just the turtle) - or just taking the picture without the menus.
    \ Screenshot: to png
\ Feature: Save and Load files
    \ Feature: Add save and restore keys -- maybe make automatic save just before playing a macro.  So can undo if press the wrong one.
    \ Feature: save and then restore, i.e. in case of a catastrophic error!  write to file.
    \ Also allows creations to be shared and a community of users to be formed.  
    \ Status bar prompts for save to file, load from file, etc.
\ Feature: Undo one instruction
\ Feature: Allow putting letters on the screen
\ Feature: View & debug macros.  Review macro recording instructions, and be able to edit... (programming and debugging instructions!)  (M mode)
\ Installer: Can DOS Box start up with the splash screen?  vDOSbox can play 32 bit files, so using IrfanView in slideshow mode as a simple screenshot.  Later - do this in Turtle Logo itself... or VGA mode, then switch back to text mode.
\ Feature: read in a saved Turtle png and essentially digitize that using turtle (as a way to load files) - basically rasterize turtle's movements using penup, pendown, and color search! :)  image processing...
\ Usability: Think through how colors should work with going forward, backspace, penup, i.e. what is painting, which square, etc.  
\ Usability: Think through which pixel should get colored, etc.  Issue is 2xF creates a square that is 3 pixels high.

\ Feature: design a random maze generator

\ Feature: Port to the Raspberry Pi, Arduino, or 8052.
    \ How big would this program be in assembly language removing all the graphics bits but just retaining the "do turtle" bits with stepper motor control?
    \ Port to Android! (keys will make it harder)  Maybe bluetooth keyboard.  Other zigbee connection with computer so one is controlling remotely with computer keyboard.
    \ Stepper motors
    \ Sensors (bumping into)
\ Feature: Port to 640x380 graphics mode screen instead of text mode (see fire.asm)
\ Feature: Re-mapping the keys to use mnemonics from a non-English language, or extend to create a multi-lingual learning environment
\ PARTIAL DONE through vDOSBox  - Feature: Port to windows via Win32 Forth
\ Feature: voice control of the turtle
\ Feature: sound feedback.  (chiptune if level reached successfully?)
\ Feature: speech synthesizer?

\ Feature: Add ability to animate, e.g. the splash screen images
\ Installer: Tell the user where the installation is going
\ Installer: test the desktop shortcut placement, and the start-menu placement.
\ Application: barack obama ascii art portrait.

\ TODO 2018-08-24--Fri--16:39 --- simplify the menu writing!!!  it is too complicated as it stands.  Use the tester code write-string
\ In progress - buggy - too many items on the stack.

\ BUG --- diary 9 is always full...
\ DONE - !! CODE REVIEW: writing text to screen or menu should be simple and uniform provide text ( ptr nchar), location, and fgbg color -- then write to menu !!
\ FEATURE --- [N/P] next / previous skill.
\ Notify if feature is full.
\ Log *all* key presses, not just when recording, so that can undo/redo/save file/load file.
\ DONE - !! CODE REVIEW: simplified string writing with uniform words for all fields. 
\ TO DO -- consolidate the fixed width string structures into a single data structure...  should it take strings from the input stream and build the structure dynamically?
\ use WORD to simplify the other string structures AND create> does...
\ USABILITY --- let the child doing recording press ESC and it terminates recording safely.  Similarly for other current forbidden actions -- i.e. don't force the child to worry about parentheses at this age :)  it's just frustrating.

\ CODE REVIEW --- search ???

\ FEATURE: [m] maze - sets up random walls creating a maze for turtle to pass through
\ FEATURE: two player turtle race

\ FIXED: 16 levels of nesting are allowed.
    \ One level nesting of recordings is allowed, but not higher level and not recursion.   LIMITATION (BUG) - only 1 level nesting is allowed??
    \ Nesting: i.e. play-macro instruction is logged if recording without playing; else play.  
    \ Recursion (not allowed): Won't log instruction to play macro when record-to diary is same current active diary
    \ First class nesting --- combining instructions and macros containing only instructions (first class macros) - ALLOWED
    \ Higher level nesting --- combining macros that themselves contain macros.

\ FEATURE: [s] save and [u] undo.
\ FEATURE: have the undo take a snapshot of itself in case of re-do...
